from enum import Enum, auto


class RuleType(Enum):
    DATA_QUALITY = auto()
    RECONCILIATION = auto()
    EQUALITY = auto()
    DATA_DRIFT = auto()
    SCHEMA_DRIFT = auto()


class RuleExecutionMode(Enum):
    SCHEDULED = auto()
    MANUAL = auto()
    WEBHOOK = auto()
    API = auto()


class RuleExecutionStatus(Enum):
    STARTED = auto()
    RUNNING = auto()
    ERRORED = auto()
    WARNING = auto()
    SUCCESSFUL = auto()
    ABORTED = auto()


class RuleResultStatus(Enum):
    STARTED = auto()
    RUNNING = auto()
    ERRORED = auto()
    WARNING = auto()
    SUCCESSFUL = auto()
    ABORTED = auto()


class PolicyExecutionType(Enum):
    SAMPLE = auto()
    FULL = auto()
    INCREMENTAL = auto()


class RuleThresholdLevel:
    def __init__(self, success, warning=None, **kwargs):
        self.success = success
        self.warning = warning
        
    def __repr__(self):
        return f"RuleThresholdLevel({self.__dict__})"


class LivyExecutorConfig:
    def __init__(self, executorMemory=None, executorCores=None, numExecutors=None, **kwargs):
        self.executorMemory = executorMemory
        self.executorCores = executorCores
        self.numExecutors = numExecutors

    def __repr__(self):
        return f"LivyExecutorConfig({self.__dict__})"


class DataBricksExecutorConfig:
    def __init__(self, minWorkers=None, maxWorkers=None, clusterWorkerType=None, clusterDriverType=None, **kwargs):
        self.minWorkers = minWorkers
        self.maxWorkers = maxWorkers
        self.clusterWorkerType = clusterWorkerType
        self.clusterDriverType = clusterDriverType
    
    def __repr__(self):
        return f"DataBricksExecutorConfig({self.__dict__})"


class ExecutorConfig:
    def __init__(self, livy=None, databricks=None, **kwargs):
        if isinstance(livy, dict):
            self.livy = LivyExecutorConfig(**livy)
        else:
            self.livy = livy
        if isinstance(databricks, dict):
            self.databricks = DataBricksExecutorConfig(**databricks)
        else:
            self.databricks = databricks

    def __repr__(self):
        return f"ExecutorConfig({self.__dict__})"


class RuleExecutionSummary:
    def __init__(self, ruleId, executionMode, executionStatus, resultStatus, startedAt, executionType, thresholdLevel,
                 ruleVersion, id=None, ruleName=None, ruleType=None, lastMarker=None, leftLastMarker=None,
                 rightLastMarker=None, executionError=None, finishedAt=None, resetPoint=None, persistencePath=None,
                 resultPersistencePath=None, executorConfig=None, markerConfig=None, leftMarkerConfig=None,
                 rightMarkerConfig=None, **kwargs):

        self.ruleId = ruleId
        self.id = id
        self.ruleName = ruleName
        if isinstance(ruleType, dict):
            self.ruleType = RuleType(**ruleType)
        else:
            self.ruleType = ruleType
        if isinstance(executionMode, dict):
            self.executionMode = RuleExecutionMode(**executionMode)
        else:
            self.executionMode = executionMode
        if isinstance(executionStatus, dict):
            self.executionStatus = RuleExecutionStatus(**executionStatus)
        else:
            self.executionStatus = executionStatus
        if isinstance(resultStatus, dict):
            self.resultStatus = RuleResultStatus(**resultStatus)
        else:
            self.resultStatus = resultStatus
        
        self.lastMarker = lastMarker
        self.leftLastMarker = leftLastMarker
        self.rightLastMarker = rightLastMarker
        self.executionError = executionError
        self.startedAt = startedAt
        self.finishedAt = finishedAt
        
        if isinstance(thresholdLevel, dict):
            self.thresholdLevel = RuleThresholdLevel(**thresholdLevel)
        else:
            self.thresholdLevel = thresholdLevel
        self.resetPoint = resetPoint
        self.persistencePath = persistencePath
        self.resultPersistencePath = resultPersistencePath
        self.ruleVersion = ruleVersion
        if isinstance(executorConfig, dict):
            self.executorConfig = ExecutorConfig(**executorConfig)
        else:
            self.executorConfig = executorConfig
        self.markerConfig = markerConfig
        self.leftMarkerConfig = leftMarkerConfig
        self.rightMarkerConfig = rightMarkerConfig
        if isinstance(executionType, dict):
            self.executionType = PolicyExecutionType(**executionType)
        else:
            self.executionType = executionType

    def __repr__(self):
        return f"RuleExecutionSummary({self.__dict__})"


class RootCauseAnalysis:
    def __init__(self, key, bad, good, badFraction, goodFraction,**kwargs):
        self.key = key
        self.bad = bad
        self.good = good
        self.badFraction = badFraction
        self.goodFraction = goodFraction

    def __repr__(self):
        return f"RootCauseAnalysis({self.__dict__})"


class RuleExecutionResult:
    def __init__(self, status, description=None, successCount=None, failureCount=None, rows=None, failedRows=None,
                 qualityScore=None, rca=None, **kwargs):
        if isinstance(status, dict):
            self.status = RuleResultStatus(**status)
        else:
            self.status = status
        self.description = description
        self.successCount = successCount
        self.failureCount = failureCount
        self.failedRows = failedRows
        self.qualityScore = qualityScore
        self.rows = rows
        if isinstance(rca, dict):
            self.rca = RootCauseAnalysis(**rca)
        else:
            self.rca = rca
    
    def __repr__(self):
        return f"RuleExecutionResult({self.__dict__})"


class RuleItemResult:
    def __init__(self, id, ruleItemId, threshold, weightage, isWarning, resultPercent=None, businessItemId=None,
                 success=None, error=None, **kwargs):
        self.id = id
        self.ruleItemId = ruleItemId
        self.businessItemId = businessItemId
        self.threshold = threshold
        self.resultPercent = resultPercent
        self.resultPercent = resultPercent
        self.success = success
        self.error = error
        self.weightage = weightage
        self.isWarning = isWarning

    def __repr__(self):
        return f"RuleItemResult({self.__dict__})"


class DataQualityRuleExecutionResult:
    def __init__(self, execution, items, meta=None, result=None, **kwargs):
        if isinstance(execution, dict):
            self.execution = RuleExecutionSummary(**execution)
        else:
            self.execution = execution
        if isinstance(result, dict):
            self.result = RuleExecutionResult(**result)
        else:
            self.result = result
        self.items = list()
        for obj in items:
            if isinstance(obj, dict):
                self.items.append(RuleItemResult(**obj))
            else:
                self.items.append(obj)

        self.meta = meta
        self.executionId = self.execution.id

    def __repr__(self):
        return f"DataQualityRuleExecutionResult({self.__dict__})"



from enum import Enum

class RuleExecutionStatus(Enum):
    STARTED = 1
    RUNNING = 2
    ERRORED = 3
    WARNING = 4
    SUCCESSFUL = 5
    ABORTED = 6


DATA_QUALITY = 'DATA_QUALITY'
RECONCILIATION = 'RECONCILIATION'



from time import sleep
import logging
from torch_sdk.errors import TorchSdkException
import torch_sdk.constants as const
LOGGER = logging.getLogger("common")


class RuleExecutionResult:
    def __init__(self, execution_details):
        self.item = execution_details.items
        self.result = execution_details.result
        self.rule_id = execution_details.execution.ruleId
        self.rule_name = execution_details.execution.ruleName
        self.rule_version = execution_details.execution.ruleVersion
        self.rule_type = execution_details.execution.ruleType
        self.started_at = execution_details.execution.startedAt
        self.finished_at = execution_details.execution.finishedAt
        self.execution_id = execution_details.execution.id

    def __eq__(self, other):
        return self.execution_id == other.execution_id

    def __repr__(self):
        return f"RuleExecutionResult({self.__dict__})"


class Executor:
    def __init__(self, rule_type, torch_client):
        self.id = None
        self.exec_call, self.result_call = get_callables(rule_type, torch_client)

    def get_status(self, execution_id=None):
        if execution_id is None:
            if self.id is not None:
                execution_id = self.id
            else:
                raise TorchSdkException('execution_id is required.')
        print('Getting rule execution status.')
        execution_details = self.result_call(execution_id=execution_id)
        return const.RuleExecutionStatus[execution_details.execution.executionStatus]

    def get_result(self, execution_id=None, sleep_interval=5, total_retries=0):
        if execution_id is None:
            if self.id is not None:
                execution_id = self.id
            else:
                raise TorchSdkException('execution_id is required.')

        retry_count = 0
        while 1:
            execution_details = self.result_call(execution_id=execution_id)
            print(f'Checking rule execution result')
            if const.RuleExecutionStatus[execution_details.execution.executionStatus] == const.RuleExecutionStatus.SUCCESSFUL \
                    and const.RuleExecutionStatus[execution_details.execution.resultStatus] == const.RuleExecutionStatus.SUCCESSFUL:
                print('Rule completed successfully.')
                return execution_details
            elif const.RuleExecutionStatus[execution_details.execution.executionStatus] == const.RuleExecutionStatus.RUNNING \
                    and const.RuleExecutionStatus[execution_details.execution.resultStatus] == const.RuleExecutionStatus.RUNNING:
                sleep(sleep_interval)
                retry_count = retry_count + 1
                if (total_retries == 0) or (total_retries > 0 and retry_count < total_retries):
                    continue
                else:
                    print(f'Exiting after {total_retries} retries.')
                    return
            else:
                raise TorchSdkException(f'Rule execution failed. Details:{execution_details}')


def get_callables(rule_type, torch_client):
    if rule_type == const.RECONCILIATION:
        result_call = torch_client.get_reconciliation_rule_result
        exec_call = torch_client.execute_reconciliation_rule
    elif rule_type == const.DATA_QUALITY:
        result_call = torch_client.get_dq_rule_result
        exec_call = torch_client.execute_dq_rule
    else:
        raise TorchSdkException(f'{rule_type} not found. Please provide correct rule_type.'
                                f' Allowed values are {const.RECONCILIATION}, {const.DATA_QUALITY}')
    return exec_call, result_call


class SyncExecutor(Executor):
    def execute(self, rule_id, incremental=False, sleep_interval=5, total_retries=0):
        execution_obj = self.exec_call(rule_id=rule_id, incremental=incremental)
        retry_count = 0
        while 1:
            execution_details = self.result_call(execution_id=execution_obj.id)
            print('Checking rule execution result.')
            if const.RuleExecutionStatus[execution_details.execution.executionStatus] == const.RuleExecutionStatus.SUCCESSFUL \
                    and const.RuleExecutionStatus[execution_details.execution.resultStatus] == const.RuleExecutionStatus.SUCCESSFUL:
                print('Rule completed successfully.')
                self.id = execution_details.execution.id
                return execution_details.execution
            elif const.RuleExecutionStatus[execution_details.execution.executionStatus] == const.RuleExecutionStatus.RUNNING \
                    and const.RuleExecutionStatus[execution_details.execution.resultStatus] == const.RuleExecutionStatus.RUNNING:
                sleep(sleep_interval)
                retry_count = retry_count + 1
                if (total_retries == 0) or (total_retries > 0 and retry_count < total_retries):
                    continue
                else:
                    print(f'Exiting after {total_retries} retries.')
            else:
                raise TorchSdkException(f'Rule execution failed. Details:{execution_details}')


class AsyncExecutor(Executor):
    def execute(self, rule_id, incremental=False):
        execution_obj = self.exec_call(rule_id=rule_id, incremental=incremental)
        self.id = execution_obj.id
        return execution_obj


